vitaut
vitaut on Hacker News
The core of newer methods like yy, xjb and zmij is remarkably simple:
https://vitaut.net/posts/2026/yy-dtoa/. Shortest uscale is basically Schubfach or, rather, it's variant called Teju Jagua and has 2-3 wide multiplications compared to 1 for newer methods.
The complexity is optional and comes from squeezing the last few nanoseconds =).
Posted by
vitaut on Tuesday August 18, 2026 @09:15AM
from the vitaut.net dept.
Posted by
vitaut on Tuesday August 11, 2026 @09:54AM
from the vitaut.net dept.
Posted by
vitaut on Friday July 17, 2026 @11:14PM
from the maurycyz.com dept.
Posted by
vitaut on Thursday May 21, 2026 @04:55PM
from the discourse.llvm.org dept.
Interviewer: Mr. Musk, I understand the wheels fell off the Cybertruck.
Musk: Well, that’s not very typical. Most vehicles are designed so the wheels don’t fall off.
Interviewer: But these ones did.
Musk: Well obviously. That’s why we recalled them. But wheel retention remains a very high priority at Tesla.
Interviewer: What caused it?
Musk: A minor component interaction that generated maximum freedom.
Interviewer: Freedom?
Musk: For the wheel.
The binary bloat is also caused by unnecessary inlining and the linker eliminates most of it (but it's still annoying e.g. for godbolt). {fmt} supports a superset of std::format and std::print features including localization. stringstream's bloat is unrelated and mostly caused by large per-call binary code from concatenation-based API.
std::print author here. Indeed, std::print shouldn't be expensive to compile, it's just a thin wrapper around a single type-erased function. The only reason why it is expensive in libstdc++ is that the type-erased function is inlined which goes against the proposed design but unfortunately can't be enforced via the standard wording and remains a Quality of Implementation (QoI) issue.
Fortunately, libstdc++ is fixing this: https://gcc.gnu.org/pipermail/gcc-patches/2026-March/710275..... There is still more work to optimize the includes but it's a good start.
Modules have been working reasonably well in clang for a while now but MSVC support is indeed buggy.
We did see build time improvements from deploying modules at Meta.
Posted by
vitaut on Wednesday January 28, 2026 @11:49AM
from the github.com dept.
The main effect of this is that some of the conversions between char and char8_t are inefficient.
Posted by
vitaut on Sunday January 25, 2026 @09:55AM
from the dougallj.wordpress.com dept.
I was impressed how fast the Rust folks adopted this! Kudos to David Tolnay and others.
Note that ~3-6ns is on modern desktop CPUs where extra few kB matter less. On microcontrollers it will be larger in absolute terms but I would expect the relative difference to also be moderate.
I don't have exact numbers but from measuring perf changes per commit it seemed that most improvements came from "printing" (e.g. switching to BCD and SIMD, branchless exponent output) and microoptimizations rather than algorithmic improvements.
If you compress the table (see my earlier comment) and use plain Schubfach then you can get really small binary size and decent perf. IIRC Dragonbox with the compressed table was ~30% slower which is a reasonable price to pay and still faster than most algorithms including Ryu.
Somewhat notable is that `char8_t` is banned with very reasonable motivation that applies to most codebases:
> Use char and unprefixed character literals. Non-UTF-8 encodings are rare enough in Chromium that the value of distinguishing them at the type level is low, and char8_t* is not interconvertible with char* (what ~all Chromium, STL, and platform-specific APIs use), so using u8 prefixes would obligate us to insert casts everywhere. If you want to declare at a type level that a block of data is string-like and not an arbitrary binary blob, prefer std::string[_view] over char*.
The shortest double-to-string algorithm is basically Schubfach or, rather, it's variation Tejú Jaguá with digit output from Dragonbox. Schubfach is a beautiful algorithm: I implemented and wrote about it in
https://vitaut.net/posts/2025/smallest-dtoa/. However, in terms of performance you can do much better nowadays. For example,
https://github.com/vitaut/zmij does 1 instead of 2-3 costly 128x64-bit multiplications in the common case and has much more efficient digit output.
It's easier to write faster code in a language with compile-time facilities such as C++ or Rust than in C. For example, doing this sort of platform-specific optimization in C is a nightmare
https://github.com/vitaut/zmij/blob/91f07497a3f6e2fb3a9f999a... (likely impossible without an external pass to generate multiple lookup tables).
Posted by
vitaut on Monday January 12, 2026 @09:18AM
from the npopov.com dept.
Please note that there is some error in your port:
Error: roundtrip fail 4.9406564584124654e-324 -> '5.e-309' -> 4.9999999999999995e-309
Error: roundtrip fail 6.6302941479442929e-310 -> '6.6302941479443e-309' -> 6.6302941479442979e-309
Error: roundtrip fail -1.9153028533493997e-310 -> '-1.9153028533494e-309' -> -1.9153028533493997e-309
Error: roundtrip fail -2.5783653320086361e-312 -> '-2.57836533201e-309' -> -2.5783653320099997e-309
Yeah, that's what I meant.
My bad, you are right. The small integer optimization should be switched to a different output method (or disabled since it doesn't provide much value). Thanks for catching this!